import typing
from abc import ABC, abstractmethod
from typing import (
    Any,
    Awaitable,
    ClassVar,
    List,
    Optional,
    Self,
    Type,
    Union,
    cast,
)

import bson
import pymongo.errors
import pymongo.results
from pydantic import BaseModel, Field, validate_call
from pymongo import MongoClient

from pydantic_mongo_document import ObjectId
from pydantic_mongo_document.config import ReplicaConfig
from pydantic_mongo_document.cursor import Cursor
from pydantic_mongo_document.document.types import InsertOneResult
from pydantic_mongo_document.document.wrapper import wrap_insert, wrap_one
from pydantic_mongo_document.encoder import JsonEncoder
from pydantic_mongo_document.exceptions import DocumentNotFound


if typing.TYPE_CHECKING:
    from motor.motor_asyncio import (
        AsyncIOMotorClient,
        AsyncIOMotorCollection,
        AsyncIOMotorDatabase,
    )
    from pymongo import MongoClient
    from pymongo.collection import Collection
    from pymongo.database import Database


CONFIG: dict[str, ReplicaConfig] = {}
"""Map of replicas to mongo URIs."""


class DocumentBase(BaseModel, ABC):
    __primary_key__: ClassVar[str] = "id"

    __replica__: ClassVar[str]
    """Mongodb replica name."""

    __database__: ClassVar[str]
    """Mongodb database name."""

    __collection__: ClassVar[str]
    """Mongodb collection name."""

    __clients__: ClassVar[dict[str, Any]] = {}
    """Map of clients for each database."""

    __document__: dict[str, Any]
    """Document data. For internal use only."""

    NotFoundError: ClassVar[Type[Exception]] = DocumentNotFound
    DuplicateKeyError: ClassVar[Type[Exception]] = pymongo.errors.DuplicateKeyError

    encoder: ClassVar[JsonEncoder] = JsonEncoder()

    id: ObjectId = Field(default_factory=lambda: str(bson.ObjectId()), alias="_id")

    def model_post_init(self, __context: Any) -> None:
        self.__document__ = self.model_dump(by_alias=True, exclude_none=True)

    @classmethod
    @abstractmethod
    def client(cls) -> Union["MongoClient", "AsyncIOMotorClient"]:
        """Returns client for database."""

    @classmethod
    @abstractmethod
    def database(cls) -> Union["Database", "AsyncIOMotorDatabase"]:
        """Returns database."""

    @classmethod
    @abstractmethod
    def collection(cls) -> Union["Collection", "AsyncIOMotorCollection"]:
        """Returns collection."""

    @property
    def primary_key(self) -> Any:
        return getattr(self, self.__primary_key__)

    @classmethod
    def get_replica_config(cls) -> ReplicaConfig:
        return CONFIG[cls.__replica__]

    @property
    def primary_key_field_name(self) -> str:
        return self.model_fields[self.__primary_key__].alias or self.__primary_key__

    @classmethod
    @validate_call
    def set_replica_config(cls, config: dict[str, ReplicaConfig]) -> None:
        CONFIG.clear()
        CONFIG.update(config)

    @classmethod
    def create_indexes(cls) -> Awaitable[None] | None:
        """Creates indexes for collection."""

    @classmethod
    @wrap_one
    def one(
        cls,
        /,
        document_id: str | None = None,
        add_query: dict[str, Any] | None = None,
        required: bool = True,
        **kwargs: Any,
    ) -> dict[str, Any] | None | Awaitable[dict[str, Any] | None]:
        """Finds one document by ID."""

        query = {}
        if document_id is not None:
            query["_id"] = document_id
        if add_query is not None:
            query.update(add_query)

        query = cls.encoder.encode_dict(query, reveal_secrets=True)

        return cast(
            dict[str, Any] | None | Awaitable[dict[str, Any] | None],
            cls.collection().find_one(query, **kwargs),
        )

    @classmethod
    def all(
        cls,
        document_ids: List[str | bson.ObjectId] | None = None,
        add_query: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Cursor[type[Self]]:
        """Finds all documents based in IDs."""

        query = {}
        if document_ids is not None:
            query["_id"] = {"$in": document_ids}
        if add_query is not None:
            query.update(add_query)

        query = cls.encoder.encode_dict(query, reveal_secrets=True)

        cursor_cls = Cursor[cls]  # type: ignore[valid-type]

        return cursor_cls(cls.collection().find(query, **kwargs))

    @classmethod
    def count(
        cls, add_query: dict[str, Any] | None = None, **kwargs: Any
    ) -> int | Awaitable[int]:
        """Counts documents in collection."""

        query = {}
        if add_query is not None:
            query.update(add_query)

        query = cls.encoder.encode_dict(query, reveal_secrets=True)

        return cast(
            int | Awaitable[int],
            cls.collection().count_documents(query, **kwargs),
        )

    def delete(
        self,
    ) -> pymongo.results.DeleteResult | Awaitable[pymongo.results.DeleteResult]:
        """Deletes document from collection."""

        query = self.encoder.encode_dict(
            {self.primary_key_field_name: self.primary_key},
        )

        return self.collection().delete_one(query)

    def commit_changes(
        self, fields: Optional[List[str]] = None
    ) -> (
        Optional[pymongo.results.UpdateResult]
        | Awaitable[Optional[pymongo.results.UpdateResult]]
    ):
        """Saves changes to document."""

        search_query: dict[str, Any] = {self.primary_key_field_name: self.primary_key}
        update_query: dict[str, Any] = {}

        if not fields:
            fields = [
                field
                for field in self.model_fields.keys()
                if field != self.__primary_key__
            ]

        data = self.encoder.encode_dict(
            self.model_dump(by_alias=True, exclude_none=True),
            reveal_secrets=True,
        )

        for field in fields:
            if field in data and data[field] != self.__document__.get(field):
                update_query.setdefault("$set", {}).update({field: data[field]})
            elif field not in data and field in self.__document__:
                update_query.setdefault("$unset", {}).update({field: ""})

        if update_query:
            return self.collection().update_one(search_query or None, update_query)

        return self.noop()

    @wrap_insert
    def insert(self) -> InsertOneResult:  # type: ignore[valid-type]
        """Inserts document into collection."""

        return cast(
            InsertOneResult,  # type: ignore[valid-type]
            self.collection().insert_one(
                self.encoder.encode_dict(
                    self.model_dump(by_alias=True, exclude_none=True),
                    reveal_secrets=True,
                )
            ),
        )

    def noop(self) -> Awaitable[None] | None:
        """No operation. Does nothing."""



from typing import Annotated, Any

from pydantic import BaseModel, Field


class ReplicaConfig(BaseModel):
    """Mongodb replica config model."""

    uri: Annotated[
        str,
        Field(..., description="Mongodb connection URI."),
    ]
    client_options: dict[str, Any] = Field(
        default_factory=dict,
        description="Mongodb client options.",
    )



import datetime
import json
import typing
from enum import Enum
from functools import cached_property
from typing import Any

import bson
import pydantic
from pydantic import BaseModel, SecretStr
from pydantic_core import Url

if typing.TYPE_CHECKING:
    from pydantic_mongo_document import ObjectId


class JsonEncoder(json.JSONEncoder):
    CONTEXT = {"document": True}

    @cached_property
    def object_id_type(self) -> pydantic.TypeAdapter["ObjectId"]:
        from pydantic_mongo_document import ObjectId

        return pydantic.TypeAdapter(ObjectId)

    def is_object_id_string(self, o: Any) -> bool:
        try:
            return self.object_id_type.validate_python(o) is not None
        except ValueError:
            return False

    def encode(self, o: Any, reveal_secrets: bool = False) -> Any:
        if isinstance(o, Enum):
            return o.value

        if isinstance(o, BaseModel):
            return o.model_dump_json(context=self.CONTEXT)

        if isinstance(o, (set, list, tuple)):
            return [self.encode(item) for item in o]

        if isinstance(o, dict):
            return self.encode_dict(o, reveal_secrets=reveal_secrets)

        if isinstance(o, SecretStr) and reveal_secrets:
            return o.get_secret_value()

        if isinstance(o, (bson.ObjectId, SecretStr, Url)):
            return str(o)

        if isinstance(o, (int, float, bool)):
            return o

        if isinstance(o, (datetime.datetime, datetime.timedelta)):
            return o

        if isinstance(o, str) and self.is_object_id_string(o):
            return bson.ObjectId(o)

        if isinstance(o, str):
            return o

        if o is None:
            return None

        return super().encode(o)

    def encode_dict(
        self, obj: dict[str, Any], reveal_secrets: bool = False
    ) -> dict[str, Any]:
        """Encodes all values in dict."""

        encoded = {}
        for key, value in obj.items():
            if isinstance(value, dict):
                encoded[key] = self.encode_dict(value, reveal_secrets=reveal_secrets)
            else:
                encoded[key] = self.encode(value, reveal_secrets=reveal_secrets)

        return encoded



import asyncio
from typing import Any, Awaitable, Self, cast

import pymongo.results

from pydantic_mongo_document.document.base import DocumentBase
from pydantic_mongo_document.document.types import AsyncInsertOneResult


from motor.motor_asyncio import (
    AsyncIOMotorClient,
    AsyncIOMotorCollection,
    AsyncIOMotorDatabase,
)


_ASYNC_CLIENTS: dict[str, AsyncIOMotorClient] = {}


class Document(DocumentBase):
    """Async document model."""

    @classmethod
    def client(cls) -> AsyncIOMotorClient:
        if cls.__replica__ not in _ASYNC_CLIENTS:
            _ASYNC_CLIENTS[cls.__replica__] = AsyncIOMotorClient(
                host=str(cls.get_replica_config().uri),
                **cls.get_replica_config().client_options,
            )

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()

        # Set the current event loop to the client's I/O loop
        _ASYNC_CLIENTS[cls.__replica__]._io_loop = loop  # noqa

        return _ASYNC_CLIENTS[cls.__replica__]

    @classmethod
    def database(cls) -> AsyncIOMotorDatabase:
        return cls.client()[cls.__database__]

    @classmethod
    def collection(cls) -> AsyncIOMotorCollection:
        return cls.database()[cls.__collection__]

    @classmethod
    async def create_indexes(cls) -> None:
        """Creates indexes for collection."""

    @classmethod
    def one(
        cls,
        document_id: str | None = None,
        add_query: dict[str, Any] | None = None,
        required: bool = True,
        **kwargs: Any,
    ) -> Awaitable[Self | None]:
        return cast(
            Awaitable[Self | None],
            super().one(document_id=document_id, add_query=add_query, **kwargs),
        )

    @classmethod
    def count(
        cls, add_query: dict[str, Any] | None = None, **kwargs: Any
    ) -> Awaitable[int]:
        return cast(Awaitable[int], super().count(add_query=add_query, **kwargs))

    def delete(
        self,
    ) -> Awaitable[pymongo.results.DeleteResult]:
        return cast(
            Awaitable[pymongo.results.DeleteResult],
            super().delete(),
        )

    def commit_changes(
        self, fields: list[str] | None = None
    ) -> Awaitable[pymongo.results.UpdateResult | None]:
        return cast(
            Awaitable[pymongo.results.UpdateResult | None],
            super().commit_changes(fields=fields),
        )

    def insert(self) -> AsyncInsertOneResult:
        return cast(AsyncInsertOneResult, super().insert())

    async def noop(self) -> None:
        """No operation. Does nothing."""



import typing
from typing import (
    Any,
    Generic,
    ParamSpec,
    Self,
    TypeVar,
    Generator,
    AsyncGenerator,
    cast,
)

import motor.motor_asyncio
import pymongo.cursor

if typing.TYPE_CHECKING:
    from pydantic_mongo_document.document.base import DocumentBase


DT = TypeVar("DT", bound=type["DocumentBase"])
P = ParamSpec("P")

AG = AsyncGenerator["DocumentBase", None]
G = Generator["DocumentBase", None, None]


class Cursor(Generic[DT]):
    model_cls: DT

    def __init__(
        self, cursor: pymongo.cursor.Cursor | motor.motor_asyncio.AsyncIOMotorCursor
    ) -> None:
        self.cursor = cursor
        self.generator: G | AG | None = None

    def __class_getitem__(cls, item: DT) -> Self:
        result = super().__class_getitem__(item)  # type: ignore[misc]
        result.model_cls = item
        return cast(Self, result)

    def __aiter__(self) -> Self:
        self.generator = self.cursor.__aiter__()
        return self

    def __iter__(self) -> Self:
        self.generator = self.cursor.__iter__()
        return self

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> "Cursor[DT]":
        return Cursor[self.model_cls](self.cursor(*args, **kwargs))  # type: ignore[name-defined]

    def __getattr__(self, item: str) -> Self | Any:
        value = getattr(self.cursor, item)

        if callable(value):
            return Cursor[self.model_cls](value)  # type: ignore[name-defined]

        return value

    async def __anext__(self) -> "DocumentBase":
        if not isinstance(self.generator, AsyncGenerator):
            raise TypeError("Cursor is not async")

        return self.model_cls.model_validate(await self.generator.__anext__())

    def __next__(self) -> "DocumentBase":
        if not isinstance(self.generator, Generator):
            raise TypeError("Cursor is not sync")

        return self.model_cls.model_validate(next(self.generator))



import asyncio
import functools
import typing
from typing import Any, Awaitable, Callable, Concatenate, ParamSpec, TypeVar

from pydantic_mongo_document.document.types import (
    AsyncInsertOneResult,
    FindOneResult,
    InsertOneResult,
    SyncInsertOneResult,
)

if typing.TYPE_CHECKING:
    from pydantic_mongo_document.document.base import DocumentBase  # noqa: F401


P = ParamSpec("P")
D = TypeVar("D", bound="DocumentBase")


def wrap_one(
    f: Callable[
        Concatenate[type[D], P],
        dict[str, Any] | None | Awaitable[dict[str, Any] | None],
    ],
) -> Callable[Concatenate[type[D], P], FindOneResult]:
    """Wraps one method to return model instance."""

    async def async_wrapper(
        cls: type[D], coro: Awaitable[dict[str, Any] | None], required: bool
    ) -> D | None:
        """Async wrapper for one method."""

        result = await coro
        if result is not None:
            return cls.model_validate(result)
        if required:
            raise cls.NotFoundError()

        return None

    def sync_wrapper(
        cls: type[D], result: dict[str, Any] | None, required: bool
    ) -> D | None:
        """Sync wrapper for one method."""

        if result is not None:
            return cls.model_validate(result)
        if required:
            raise cls.NotFoundError()

        return None

    def wrapper(
        cls: type[D], /, *args: P.args, **kwargs: P.kwargs
    ) -> D | None | Awaitable[D | None]:
        """Wrapper for one method."""

        result = f(cls, *args, **kwargs)
        required: bool = bool(kwargs.get("required", True))

        return (
            async_wrapper(cls, result, required)
            if isinstance(result, Awaitable)
            else sync_wrapper(cls, result, required)
        )

    return wrapper


def wrap_insert(
    f: Callable[Concatenate[D, P], InsertOneResult],  # type: ignore[valid-type]
) -> Callable[Concatenate[D, P], D | Awaitable[D]]:
    """Wraps insert method to return model instance."""

    async def async_wrapper(self: D, result: AsyncInsertOneResult) -> D:
        """Async wrapper for insert method."""

        result = await result

        if getattr(self, self.__primary_key__, None) is None:
            setattr(self, self.__primary_key__, result.inserted_id)

        return self

    def sync_wrapper(self: D, result: SyncInsertOneResult) -> D:  # type: ignore[valid-type]
        """Sync wrapper for insert method."""

        if getattr(self, self.__primary_key__, None) is None:
            setattr(self, self.__primary_key__, result.inserted_id)  # type: ignore[attr-defined]

        return self

    @functools.wraps(f)
    def wrapper(self: D, /, *args: P.args, **kwargs: P.kwargs) -> D | Awaitable[D]:
        """Wrapper for insert method."""

        result = f(self, *args, **kwargs)

        return (
            async_wrapper(self, result)
            if asyncio.iscoroutine(result)
            else sync_wrapper(self, result)
        )

    return wrapper



from typing import Annotated, Literal, TypeVar

import bson
import bson.json_util
from pydantic import BeforeValidator, StringConstraints

T = TypeVar("T")

_ObjectIdString = Annotated[
    str, StringConstraints(min_length=24, max_length=24, pattern=r"^[a-f\d]{24}$")
]

_DictObjectId = dict[Literal["$oid"], _ObjectIdString]


def check_object_id(value: _ObjectIdString | _DictObjectId) -> str:
    if isinstance(value, dict):
        value = value["$oid"]

    if not bson.ObjectId.is_valid(value):
        raise ValueError("Invalid ObjectId")

    return str(value)


ObjectId = Annotated[
    _ObjectIdString,
    BeforeValidator(check_object_id),
]

__all__ = ["ObjectId"]



from typing import Any, List, Optional, Self, cast

from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.mongo_client import MongoClient

from pydantic_mongo_document.document.base import DocumentBase

import pymongo.results

_SYNC_CLIENTS: dict[str, MongoClient] = {}


class Document(DocumentBase):
    """Sync document model."""

    @classmethod
    def client(cls) -> "MongoClient[Self]":
        if cls.__replica__ not in _SYNC_CLIENTS:
            _SYNC_CLIENTS[cls.__replica__] = MongoClient(
                str(cls.get_replica_config().uri),
                **cls.get_replica_config().client_options,
            )

        return _SYNC_CLIENTS[cls.__replica__]

    @classmethod
    def database(cls) -> "Database[Self]":
        return cls.client()[cls.__database__]

    @classmethod
    def collection(cls) -> "Collection[Any]":
        return cls.database()[cls.__collection__]

    @classmethod
    def create_indexes(cls) -> None:
        """Creates indexes for collection."""

    @classmethod
    def one(
        cls,
        document_id: str | None = None,
        add_query: dict[str, Any] | None = None,
        required: bool = True,
        **kwargs: Any,
    ) -> Self | None:
        return cast(
            Self | None,
            super().one(document_id=document_id, add_query=add_query, **kwargs),
        )

    @classmethod
    def count(cls, add_query: dict[str, Any] | None = None, **kwargs: Any) -> int:
        return cast(int, super().count(add_query=add_query, **kwargs))

    def insert(self) -> Self:
        return cast(Self, super().insert())

    def noop(self) -> None:
        return

    def commit_changes(
        self, fields: Optional[List[str]] = None
    ) -> Optional[pymongo.results.UpdateResult]:
        return cast(
            Optional[pymongo.results.UpdateResult],
            super().commit_changes(fields=fields),
        )

    def delete(
        self,
    ) -> pymongo.results.DeleteResult:
        return cast(
            pymongo.results.DeleteResult,
            super().delete(),
        )



from . import compat  # noqa:F401 # isort:skip
from . import document
from .config import ReplicaConfig
from .encoder import JsonEncoder
from .exceptions import DocumentNotFound
from .types import ObjectId

__all__ = [
    "document",
    "DocumentNotFound",
    "JsonEncoder",
    "ObjectId",
    "ReplicaConfig",
]



import typing
from typing import Awaitable, TypeVar, Optional
import pymongo.results

if typing.TYPE_CHECKING:
    from pydantic_mongo_document.document.base import DocumentBase  # noqa: F401


D = TypeVar("D", bound="DocumentBase")

SyncInsertOneResult = pymongo.results.InsertOneResult
AsyncInsertOneResult = Awaitable[SyncInsertOneResult]  # type: ignore[valid-type]
InsertOneResult = SyncInsertOneResult | AsyncInsertOneResult
SyncFindOneResult = Optional["DocumentBase"]
AsyncFindOneResult = Awaitable[SyncFindOneResult]
FindOneResult = SyncFindOneResult | AsyncFindOneResult
